
use(channels).


/* AMULET3 re-order buffer model */

/* This version doesn't do conditional instructions and has only one
   allocate and one forwarding port. */
   
   
/* The re-order buffer (ROB) has six ports:
   - result receives results from the execute pipeline
   - fwd_req and fwd_response receive forwarding requests and send
     responses to the start of the execute pipeline.
   - reg_write sends updates to the register bank.
   - alloc_req and alloc_response receive and respond to requests to
     allocate entries from the issuer. */

/* This is the type of results received from the execute pipeline: */

result_type:type=record(tag::int,data::int).

        /* tag is the tag for this result returned by some previous
           allocation request. */
           
        /* data is the actual result value. */

/* This is the type of responses to forwarding requests: */

fwd_response_type:type=record(found::bool,data::int).

        /* found is a boolean indicating whether a value for
           forwarding was found. */
        
        /* data is the required data value, if found. */
        
/* This is the type of write requests to the register bank: */

reg_write_type:type=record(regnum::int,data::int).

        /* regnum indicates the register number to be written to */
        
        /* data indicates the data value to be written */


/* The ROB model starts here */

/* The number of entries in the re-order buffer (ROB) is adjustable 
   using the rob_size parameter. */

rob(rob_size:val(int),
    result1:var(chan(result_type)),
    result2:var(chan(result_type)),
    fwd_req:var(chan(int)),
    fwd_response:var(chan(fwd_response_type)),
    reg_write:var(chan(reg_write_type)),
    alloc_req:var(chan(int)),
    alloc_response:var(chan(int))) : expr(void) = (
     
  /* Central to the model is a content-addressable memory (CAM): */
  
  cam_entry_type:type=record(regnum::int,
                             data::int,
                             valid::bool,
                             written_back::bool).
  
          /* regnum is the number of the register that this entry
             coresponds to, or -1 if there is no coresponding register. */
          
          /* data is the data value, if known. */
          
          /* valid indicates whether the data value is known. */
  
          /* written_back indicates whether the value has been written
             into the register bank. */
  
  /* The cam itself: */
  
  cam:var(cam_entry_type^rob_size).
  
  /* The cam must be initialised empty: */
  
  init_cam:expr(void)=(
      i:var(int).
      
      forseq i in 0 to rob_size-1 do (
          (cam[i])->regnum:=-1;
          (cam[i])->valid:=false;
          (cam[i])->written_back:=true
      )
  ).
  
  /* We need to keep track of which of the cam entries contains the most
     recently allocated value: */
     
  most_recent_entry:var(int).
  
  
  /* The rob contains four processes: */
  
  
  /* This process waits for a result from the execution unit and stores
     it in the cam entry given by its tag: */
     
  update(result:var(chan(result_type))):expr(void)=(
      forever(
        result?(
            wait_for(50);
            (cam[(?result)->tag])->data:=(?result)->data;
            (cam[(?result)->tag])->valid:=true
        )
      )
  ).
  
  
  /* This process waits for a forwarding request and supplies any
     coresponding data. */
     
  forward(fwd_req:var(chan(int)),
          fwd_response:var(chan(fwd_response_type))):expr(void)=(
      i:var(int).
      entry:var(int).
      done:var(bool).
      found:var(bool).
      response:var(fwd_response_type).
      regnum:var(int).
      
      forever(
        fwd_req?(
            regnum:=(?fwd_req);
            /* Search through the cam, from most recent to least recent entry,
               looking for the first entry whose register number matches the
               register number in this request (if any). */
            i:=most_recent_entry;
            done:=false;
            while !done do (
              if ((cam[i])->regnum=regnum) then (
                entry:=i;
                found:=true;
                done:=true
              ) else (
                i:=(i+rob_size-1) mod rob_size;
                if (i=most_recent_entry) then (
                  found:=false;
                  done:=true
                )
              )
            );
            wait_for(60);
            if found then (
                /* The requested register is in the cam, so when that cam
                   entry contains valid data, return it. */
                wait_until((cam[entry])->valid);
                response->found:=true;
                response->data:=(cam[entry])->data;
                fwd_response!response
            ) else (
                /* The requested register is not in the cam, so send a
                   negative response. */
                response->found:=false;
                fwd_response!response
            )
        )
      )
  ).
  
  
  /* This process writes results to the register bank */
  
  write_back(reg_write:var(chan(reg_write_type))):expr(void)=(
      rw:var(reg_write_type).
      entry:var(int).
      
      forever(
        forseq entry in 0 to rob_size-1 do (
            wait_until((cam[entry])->valid 
                       && (!(cam[entry])->written_back));
            wait_for(30);
            rw->regnum:=(cam[entry])->regnum;
            rw->data:=(cam[entry])->data;
            reg_write!rw;
            (cam[entry])->written_back:=true
        )
      )
  ).
  
  
  /* This process handles allocate requests */
  
  alloc(alloc_req:var(chan(int)),alloc_response:var(chan(int))):expr(void)=(
      entry:var(int).
      
      forever(
        forseq entry in 0 to rob_size-1 do (
            alloc_req?(
                wait_until((cam[entry])->written_back);
                wait_for(30);
                (cam[entry])->regnum:=(?alloc_req);
                (cam[entry])->valid:=false;
                (cam[entry])->written_back:=false;
                most_recent_entry:=entry;
                alloc_response!entry
            )
        )
      )
  ).


  /* Here is the body of the rob: */

  init_cam;
  most_recent_entry:=0;
  (
    update(result1) |
    update(result2) |
    forward(fwd_req,fwd_response) |
    write_back(reg_write) |
    alloc(alloc_req,alloc_response)
  )
).



/* Here is a test harness: */

alu_result:var(chan(result_type)).
load_result:var(chan(result_type)).
load_req:var(chan(result_type)).
fwd_req:var(chan(int)).
fwd_response:var(chan(fwd_response_type)).
reg_write:var(chan(reg_write_type)).
alloc_req:var(chan(int)).
alloc_response:var(chan(int)).

init_chan(alu_result,"alu_result");
init_chan(load_result,"load_result");
init_chan(load_req,"load_req");
init_chan(fwd_req,"fwd_req");
init_chan(fwd_response,"fwd_response");
init_chan(reg_write,"reg_write");
init_chan(alloc_req,"alloc_req");
init_chan(alloc_response,"alloc_response");

(
  rob(4,alu_result,load_result,fwd_req,fwd_response,reg_write,alloc_req,
      alloc_response) |
  
  (
    "issuer" is (         /* model instruction issuer */
      src:var(int).
      dest:var(int).
      
      forever(
        src:=rand(16);
        dest:=rand(16);
/*        msg("src="@i2s(src)@", dest="@i2s(dest)); */
        print(i2s(now),": issuer: src=",i2s(src),", dest=",i2s(dest),"\n");
        wait_for(20);
        fwd_req!src;
        alloc_req!dest;
        wait_for(60)
      )
    )
  ) |
  (
    "execute" is (       /* model execute path */
      tag:var(int).
      res:var(result_type).
      fwdresp:var(fwd_response_type).
      
      forever(
        (
          alloc_response?(
              tag:=(?alloc_response)
          ) |
          fwd_response?(
              fwdresp:=(?fwd_response);
              print(i2s(now),
                (if (fwdresp->found)
                 then ": execute: forwarded:"
                 else ": execute: not forwarded:"))
          )
        );              
        res->tag:=tag;
        res->data:=rand(100);
        print(" tag=",i2s(tag),", data=",i2s(res->data),"\n");
        wait_for(70);
        if (rand(4)=1) then
          load_req!res
        else
          alu_result!res
      )
    )
  ) |
  (
    "memory" is (        /* model data memory */
      x:var(result_type).
      forever(
        load_req?(x:=(?load_req));
        wait_for(200);
        load_result!x
      )
    )
  ) |
  (
    "regbank" is (       /* model register bank */
      rw:var(reg_write_type).

      forever(
        reg_write?(
            rw:=(?reg_write);
            print(i2s(now),": reg_write: reg=",i2s(rw->regnum),
                  ", data=",i2s(rw->data),"\n");
            wait_for(50)
        )
      )
    )
  )
)

